#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Chi2.h"
#include "opencl/simpleCL.h"

int main(int argc ,char* argv[]) {

FILE *fp;
size_t size;
char OPENCL_KERNEL_FILE[1000]; OPENCL_KERNEL_FILE[0] = '\0';
	  
/* Initialize rows, cols, ncases, ncontrols from the user */
unsigned int rows=atoi(argv[2]);
unsigned int cols=atoi(argv[3]);
int ncases=atoi(argv[4]);
int ncontrols=atoi(argv[5]);
float bcutoff = atof(argv[6]);
strcpy(OPENCL_KERNEL_FILE, argv[7]);
int CUDA_DEVICE = atoi(argv[8]);
printf("Individuals=%d SNPs=%d cases=%d controls=%d Bonferroni-cutoff=%f CUDA_DEVICE=%d KERNEL=%s\n",rows,cols,ncases,ncontrols,bcutoff,CUDA_DEVICE, OPENCL_KERNEL_FILE);

/*Host variable declaration */

int THREADS = 256;
int BLOCKS;
float host_results[cols];
unsigned int jobs; 
int transfer = 0;
unsigned long totaljobs ;
unsigned long ulone = 1;
unsigned long ultwo = 2; 

sclHard* hardware; sclSoft software; int found;
cl_mem dev_dataT_clmem, results_clmem;
size_t global_work_size[2], local_work_size[2];

/*Kernel variable declaration */
unsigned char *dev_dataT;
float *results;

/* set the cut off */
float cutoff = 0.05/cols;
printf("Univariate cutoff is %e\n",cutoff);

/* Validation to check if the data file is readable */
fp = fopen(argv[1], "r");
if (fp == NULL) {
    	printf("Cannot Open the File");
	return 0;
}

size = (size_t)(rows * cols);
totaljobs = cols * (cols - ulone )/ ultwo ;
printf("Totaljobs is %lu\n",totaljobs);
printf("Size of the data = %u\n",size);

// Get the hardware
hardware = sclGetAllHardware(&found);
sclPrintHardwareStatus(*hardware);

// Get software
software = sclGetCLSoftware(OPENCL_KERNEL_FILE, "chi2_kernel", hardware[0]);

fflush(stdout);

unsigned char *dataT = (unsigned char*)malloc((size_t)size);

if(dataT == NULL) {
        printf("ERROR: Memory for data not allocated.\n");
}

/* Transfer the SNP Data from the file to CPU Memory */
unsigned int i;
for(i=0 ; i < size; i++ ) {
	int tmp;
	fscanf(fp,"%d",&tmp);
	dataT[i] = (char)(((int)'0')+tmp);
	if(i == size - 1){ 
		transfer = 1; 
		printf("SNP Data Transferred to the Memory....\nProcessing\n");
   	}
} 
fclose(fp);

if(transfer != 1){
        printf("ERROR: Complete data is not transferred to the Memory.\n");
}

/* allocate the Memory in the GPU for SNP data */	   
//cudaMalloc((unsigned char**) &dev_dataT, size * sizeof(unsigned char) );
dev_dataT_clmem = sclMalloc( hardware[0], CL_MEM_READ_WRITE, size * sizeof(unsigned char) );

//cudaMalloc((float**) &results, cols * sizeof(float) );
results_clmem = sclMalloc( hardware[0], CL_MEM_READ_WRITE, cols * sizeof(float) );

/*Copy the SNP data to GPU */
//cudaMemcpy(dev_dataT, dataT, size * sizeof(unsigned char), cudaMemcpyHostToDevice);
sclWrite( hardware[0], size * sizeof(unsigned char), dev_dataT_clmem, (void*) dataT );

printf("SNP 1   Uni p-value     SNP 2   Uni p-value     Chi-Square      P-value\n");
printf("#####   ###########     #####   ###########     ########        ########\n");

jobs = cols;
BLOCKS = (jobs + THREADS - 1)/THREADS;

global_work_size[0] = BLOCKS * THREADS;
global_work_size[1] = 1;
local_work_size[0] = THREADS;
local_work_size[1] = 1;

/*Calling the kernel function */
//kernel<<<BLOCKS,THREADS>>>(rows,cols,ncases,ncontrols,dev_dataT,results);

sclSetKernelArg( software, 0, sizeof(int), (void*) &rows);
sclSetKernelArg( software, 1, sizeof(int), (void*) &cols);
sclSetKernelArg( software, 2, sizeof(int), (void*) &ncases);
sclSetKernelArg( software, 3, sizeof(int), (void*) &ncontrols);
sclSetKernelArg( software, 4, sizeof(cl_mem), (void*) &dev_dataT_clmem);
sclSetKernelArg( software, 5, sizeof(cl_mem), (void*) &results_clmem);

sclLaunchKernel( hardware[0], software, global_work_size, local_work_size );
		
/*Copy the results back in host*/
//cudaMemcpy(host_results,results,cols * sizeof(float),cudaMemcpyDeviceToHost);
sclRead( hardware[0], cols * sizeof(float), results_clmem, host_results );

//cudaMemcpy(results, dev_results, JOBS * sizeof(float), cudaMemcpyDeviceToHost);
//clRead( hardware[0], (size_t)(JOBS * sizeof(float)), dev_results_clmem, results );

int k;		
for(k = 0; k < jobs; k++) {
	printf("%f ", host_results[k]);
}
printf("\n");

//cudaFree( dev_dataT );
//cudaFree( results );
return 0;

}
